/*------------------------------------------------------------------------
   OO    O   OOOOO   OOOO   O    O     The North Carolina State
   O O   O  O       O       O    O     University Motif Group
   O  O  O  O        OOOO   O    O     NCSU Box 8206
   O   O O  O            O  O    O     Raleigh, NC 2695-8206
   O    OO   OOOOO   OOOO    OOOO

   O     O  OOOO  OOOOOOO O  OOOOO      OOOO  OOOO   OOO  O    O OOOO
   OO   OO O    O    O    O  O         O    O O   O O   O O    O O   O
   O O O O O    O    O    O  OOOOO     O      OOOO  O   O O    O OOOO
   O  O  O O    O    O    O  O         O  OOO O O   O   O O    O O
   O     O O    O    O    O  O         O    O O  O  O   O O    O O 
   O     O  OOOO     O    O  O          OOOO  O   O  OOO   OOOO  O
 ------------------------------------------------------------------------

   Title: NMG Editor demo

   File name: nmg_editor1.c

   File Type: Code

   Lead Programmer: Marshall Brain, brain@eos.ncsu.edu

   Testing by: Errol Casey, Dave Patterson, Rob Ward, Kelly Campbell
               Brian Casper, Lance Lovette, Aaron Nauman, Steve Loyer,
               Jerry Cox.

   Version: 1.03
   Date   : 11/5/91

   Description: This program implements a Motif text editor. It
        includes File open, new, save, save as, close and quit; 
        Clipboard cut, copy, paste and clear; and Navigate top, 
        bottom, jump to cursor and jump to line. It also includes
        a utility menu containing a find dialog. Use it to learn 
        about the text widget, or to create a personalized editor.

   Author's notes:
        I have written a book on Motif programming (due from Digital
        Press in early 1992). This program was written to prove some
        of the code in the book correct. For detailed descriptions
        of the code, please see the book (all chapter references in
        the comments refer to chapters in the book). To order the book
        call 1-800-DIGITAL and ask for EY-J816E-P.

        The find dialog used in this program will be more useful if 
        it has a title bar that allows it to be moved. If dialog boxes 
        on your system do not have title bars and other decorations, then 
        add the following line to your ".Xdefaults" file:
              mwm*transientDecoration:   all

        Occasionally the scroll bar in the text widget will not allow
        you to scroll all the way to the bottom of the text. I believe
        this is a problem inside the text widget. To solve this, use
        the Navigate/Bottom menu option. It will go to the bottom of
        the text and reset the scroll bar. 

   System Information:
        This text and code is based on the OSF/Motif widget set version
        1.1 and the X Window System version 4.

   About the NCSU Motif Group:
        The NCSU Motif Group is a group of students at North Carolina
        State University who meet once a week to discuss Motif programming.
        Individual projects are used to help students learn Motif. 
        These projects are then released onto the network to help others.
        All released code is available at our anonymous FTP site:
        osl.csc.ncsu.edu, in the directory pub/ncsu_motif.  Questions 
        and comments should be directed to the lead programmer for the 
        project in question, or to the faculty advisor, Marshall Brain, 
        at brain@eos.ncsu.edu.

   Copyright Information:
        Copyright 1991 by Marshall Brain
        All Rights Reserved

        Permission to use, copy, and distribute this software and text for
        non-commercial purposes and without fee is hereby granted, provided 
        that this notice appears in all copies.

        The author disclaims all warranties with regard to the software or 
        text including all implied warranties of merchantability and fitness.

        In no event shall the author or NCSU be liable for any special,
        indirect or cosequential damages or any damages whatsoever 
        resulting from loss of use, data or profits, whether in an 
        action of contract, negligence or other tortious action, 
        arising out of or in connection with the use or performance 
        of this software or text.

***************************************************************************/

#include <stdio.h>
#include <sys/types.h> 
#include <sys/stat.h> 

#include <X11/cursorfont.h>

#include <Xm/Xm.h>
#include <Xm/Text.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
#include <Xm/CascadeB.h>
#include <Xm/FileSB.h>
#include <Xm/SelectioB.h>
#include <Xm/MessageB.h>
#include <Xm/BulletinB.h>
#include <Xm/ToggleB.h>

/* constants for menu options and find dialog buttons. */
#define OPEN        11
#define CLOSE       12
#define NEW         13
#define SAVE        14
#define SAVE_AS     16
#define QUIT        17

#define CUT         21
#define CLEAR       22
#define COPY        23
#define PASTE       24

#define TOP         35
#define BOTTOM      36
#define JUMP_CURS   37
#define JUMP_LINE   38

#define ABOUT       41
#define HELP        42

#define FIND        51

#define OK          1
#define CANCEL      2

#define FIND_FIND        1
#define FIND_FIND_CHANGE 2
#define FIND_CHANGE      3
#define FIND_CANCEL      4
#define FIND_CASE        5

/* bitmap for icon */
#define nmg_width 50
#define nmg_height 50
static char nmg_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0e, 0x98, 0x03, 0x0e, 0x7e,
   0x00, 0x00, 0x1e, 0x98, 0x07, 0x8f, 0xff, 0x00, 0x00, 0x1e, 0x98, 0x07,
   0x8f, 0x81, 0x01, 0x00, 0x36, 0x98, 0x8d, 0xcd, 0x80, 0x01, 0x00, 0x36,
   0x98, 0x8d, 0xcd, 0x00, 0x00, 0x00, 0x66, 0x98, 0xd9, 0xcc, 0x00, 0x00,
   0x00, 0x66, 0x98, 0xd9, 0xcc, 0x00, 0x00, 0x00, 0xc6, 0x98, 0x71, 0xcc,
   0x00, 0x00, 0x00, 0xc6, 0x98, 0x71, 0xcc, 0xf8, 0x01, 0x00, 0x86, 0x99,
   0x01, 0xcc, 0xf8, 0x01, 0x00, 0x86, 0x99, 0x01, 0xcc, 0x80, 0x01, 0x00,
   0x06, 0x9b, 0x01, 0xcc, 0x80, 0x01, 0x00, 0x06, 0x9b, 0x01, 0xcc, 0x80,
   0x01, 0x00, 0x06, 0x9e, 0x01, 0x8c, 0x81, 0x01, 0x00, 0x06, 0x9c, 0x01,
   0x8c, 0xff, 0x00, 0x00, 0x06, 0x9c, 0x01, 0x0c, 0x7e, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x03, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x42, 0x8e, 0x13, 0x01, 0x00, 0x00, 0x00,
   0x46, 0x51, 0x14, 0x01, 0x00, 0x00, 0x00, 0x4a, 0x41, 0x10, 0x01, 0x00,
   0x00, 0x00, 0x4a, 0x81, 0x11, 0x01, 0x00, 0x00, 0x00, 0x52, 0x01, 0x12,
   0x01, 0x00, 0x00, 0x00, 0x52, 0x01, 0x14, 0x01, 0x00, 0x00, 0x00, 0x62,
   0x51, 0x14, 0x01, 0x00, 0x00, 0x00, 0x42, 0x8e, 0xe3, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x22, 0xe7, 0xeb, 0x03, 0x00, 0x00, 0x00, 0xb6, 0x88,
   0x28, 0x00, 0x00, 0x00, 0x00, 0xaa, 0x88, 0x28, 0x00, 0x00, 0x00, 0x00,
   0xa2, 0x88, 0xe8, 0x00, 0x00, 0x00, 0x00, 0xa2, 0x88, 0x28, 0x00, 0x00,
   0x00, 0x00, 0xa2, 0x88, 0x28, 0x00, 0x00, 0x00, 0x00, 0x22, 0x87, 0x28,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x9c, 0xc7, 0x89, 0x1e, 0x00, 0x00,
   0x00, 0xa2, 0x28, 0x8a, 0x22, 0x00, 0x00, 0x00, 0x82, 0x28, 0x8a, 0x22,
   0x00, 0x00, 0x00, 0xba, 0x27, 0x8a, 0x1e, 0x00, 0x00, 0x00, 0xa2, 0x22,
   0x8a, 0x02, 0x00, 0x00, 0x00, 0xa2, 0x24, 0x8a, 0x02, 0x00, 0x00, 0x00,
   0x9c, 0xc8, 0x71, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00};


/* general widgets. */
Widget toplevel, text, form, menu_bar;

/* dialog box widgets */
Widget open_dialog, new_dialog, jump_dialog, quit_dialog, 
       readonly_dialog, error_dialog, close_dialog, save_as_dialog,
       about_dialog, help_dialog, find_dialog, finderror_dialog,
       overwrite_dialog;

/* widgets having to do with find dialog */
Widget find_label1, find_label2;
Widget find_edit1, find_edit2;
Widget find_button, find_change_button, change_button, find_top_button,
    cancel_button;
Widget case_toggle;

/* menu item widgets. */
Widget open_option, new_option, save_option, save_as_option,
       close_option, quit_option;
Widget cut_option, clear_option, copy_option, paste_option;
Widget top_option,bottom_option, jump_curs_option, jump_line_option;
Widget help_option, about_option;
Widget find_option, find_rc;

XtAppContext context;
XmStringCharSet char_set=XmSTRING_DEFAULT_CHARSET;

char *filename=NULL;         /* holds the current filename */
char *new_filename=NULL;     /* holds new filename for save as option. */
Boolean text_changed=False;  /* tells whether file has been modified. */
Boolean case_matters=False;  /* remembers case sensitivity for find. */

void change_menu_sensitivity(open_state)
    Boolean open_state;
/* changes the menu sensitivities between open and close states. */
{
    XtSetSensitive(open_option,open_state);
    XtSetSensitive(new_option,open_state);
    XtSetSensitive(close_option,!open_state);
    XtSetSensitive(save_option,False);          /* will get set true if */
    XtSetSensitive(save_as_option,!open_state); /* text is modified.    */

    XtSetSensitive(cut_option,!open_state);
    XtSetSensitive(copy_option,!open_state);
    XtSetSensitive(paste_option,!open_state);
    XtSetSensitive(clear_option,!open_state);

    XtSetSensitive(top_option,!open_state);
    XtSetSensitive(bottom_option,!open_state);
    XtSetSensitive(jump_curs_option,!open_state);
    XtSetSensitive(jump_line_option,!open_state);

    XtSetSensitive(find_option,!open_state);
    XtSetSensitive(find_change_button,!open_state);
    XtSetSensitive(change_button,!open_state);
}

void watch_cursor(w)
    Widget w;
/* change the cursor to a wrist watch shape. */
/* See Chapter 15. */
{
    Cursor c1;

    c1 = XCreateFontCursor(XtDisplay(w),XC_watch);
    XDefineCursor(XtDisplay(w),XtWindow(w),c1);
    XFlush(XtDisplay(w));
}

void normal_cursor(w)
    Widget w;
/* return the cursor to its normal shape. */
/* See Chapter 15. */
{
    XUndefineCursor(XtDisplay(w),XtWindow(w));
    XFlush(XtDisplay(w));
}

void change_title()
/* changes the title on the window to file name. */
/* See Chapter 14. */
{
    int ac;
    Arg al[10];
    char s[1000];

    if (filename!=NULL)
    {
        strcpy(s,"NMG Editor - ");
        strcat(s,filename);
        ac = 0;
        XtSetArg(al[ac], XmNtitle, s); ac++;
        XtSetValues(toplevel,al,ac);
    }
    else
    {
        strcpy(s,"NMG Editor");
        ac = 0;
        XtSetArg(al[ac], XmNtitle, s); ac++;
        XtSetValues(toplevel,al,ac);
    }
}

Boolean read_file()
/* reads the file in filename into the text widget. */
/* See Chapter 10. */
{
    FILE *f;
    char *file_contents;
    int file_length;
    struct stat stat_val;

    /* open and read the file. */
    if (stat(filename, &stat_val) == 0)
    {
        watch_cursor(toplevel);
        file_length = stat_val.st_size;
        /* try to open file in "r+" mode. if OK then read it. */
        if ((f=fopen(filename,"r+"))==NULL)
        {
            /* if can't open with "r+", try to open with "r". This 
               means the file is read-only. If can't open that way, 
               then file is unreadable. */
            if ((f=fopen(filename,"r"))==NULL)
                return(False);
            else
                XtManageChild(readonly_dialog);
        }
        /* malloc a place for the string to be read to. */
        file_contents = (char *) XtMalloc(file_length+1);
        *file_contents = '\0';
        /* read the file string */
        fread(file_contents, sizeof(char), file_length, f);
        fclose(f);
        file_contents[file_length]='\0';
        /* give the string to the text widget. */
        XmTextSetString(text, file_contents);
        XtFree(file_contents);
        text_changed=False;
        XtSetSensitive(save_option,False);
        normal_cursor(toplevel);
    }
    else
        return(False);
    return(True);
}

void openCB(w, client_data, call_data)
    Widget w;
    int client_data;
    XmFileSelectionBoxCallbackStruct *call_data;
/* handles the file selection box callback and the new dialog callback. 
   if the file is new, then read_file will do nothing and a blank
   file will be created. */
/* See Chapter 10. */
{
    if (client_data==CANCEL) /* do nothing if cancel is selected. */
    {
        XtUnmanageChild(open_dialog);
        return;
    }

    if (filename != NULL) /* free up filename if it exists. */
    {
        XtFree(filename);
        filename = NULL;
    }

    /* get the filename from the file selection box */
    XmStringGetLtoR(call_data->value, char_set, &filename);
    change_title();

    if (read_file()||(w==new_dialog))
    {
        change_menu_sensitivity(False);
        XtSetSensitive(text,True);
        XmTextSetEditable(text,True);
        XmTextSetInsertionPosition(text,0);
        XtUnmanageChild(open_dialog);
    }
}

void handle_close()
/* closes the file and returns text widget to blank state. */
/* See Chapter 10. */
{
    int ac;
    Arg al[10];

    XtSetSensitive(text,False);
    XmTextSetEditable(text,False);
    XmTextSetString(text,"");
    change_menu_sensitivity(True);
    text_changed=False;

    /* change title back to default. */
    ac=0;
    XtSetArg(al[ac], XmNtitle, "NMG Editor"); ac++;
    XtSetValues(toplevel,al,ac);
}

Boolean handle_save(filename)
    char *filename;
/* saves the text widget's string to a file. */
/* See Chapter 10. */
{
    FILE *f;
    char *s=NULL;
   
    /* otherwise, prepare to write the file. If cannot write the file,
       display an error dialog. */
    if ((f=fopen(filename,"w"))!=NULL)
    {
        watch_cursor(toplevel);
        /* get the string from the text widget */
        s = (char *)XmTextGetString(text);
        if (s!=NULL) 
        {
            /* write the file. */
            fwrite(s, sizeof(char), strlen(s), f);
            /* make sure the last line is terminated by '\n' 
               so that vi, compilers, etc. like it. */
            if (s[strlen(s)-1]!='\n')
                fprintf(f,"\n");
            XtFree(s); 
        }
        fflush(f);
        fclose(f);
        text_changed=False;
        normal_cursor(toplevel);
        return True;
    }
    else
    {
        XtManageChild(error_dialog);
        return False;
    }
}

void text_changedCB(w, client_data, call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
/* Each time text widget value is changed, this CB is called. 
   It allows the quit option to detect that the text has been changed.*/
/* See Chapter 10. */
{
    if (!text_changed)
    {
        XtSetSensitive(save_option,True);
        text_changed=True;
    }
}

void overwriteCB(w, client_data, call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
/* handles the "ok to overwrite" dialog for save as. */
{
    if (client_data==CANCEL)
    {
        XtUnmanageChild (overwrite_dialog);
        return;
    }
    else if (handle_save(new_filename))
    {
        if (filename != NULL) /* free up filename if it exists. */
        {
            XtFree(filename);
            filename = NULL;
        }
        filename=new_filename;
        change_title();
        XtSetSensitive(save_option,False);    /* will get set true if  */
                                              /* changed again. */
    }
}

void save_asCB(w, client_data, call_data)
    Widget w;
    int client_data;
    XmSelectionBoxCallbackStruct *call_data;
/* handles retrieval of new file name dialog box. */
{
    FILE *f;

    if (client_data==CANCEL)
    {
        XtUnmanageChild (save_as_dialog);
        return;
    }
        
    /* get the filename from the dialog box */
    XmStringGetLtoR(call_data->value, char_set, 
        &new_filename);

    if ((f=fopen(new_filename,"r"))!=NULL)
    {
        fclose(f);
        XtManageChild(overwrite_dialog);
    }
    else if (handle_save(new_filename))
    {
        if (filename != NULL) /* free up filename if it exists. */
        {
            XtFree(filename);
            filename = NULL;
        }
        filename=new_filename;
        change_title();
        XtSetSensitive(save_option,False);    /* will get set true if  */
                                              /* changed again. */
    }
}

void jump_to_line(line_num)
    int line_num;
/* Counts '\n's so that cursor can be placed at correct line. */
/* See Chapter 10. */
{
    int x,l,curr;
    char *temp;
    Arg al[10];
    int ac;

    /* get string from text widget */
    temp=XmTextGetString(text);
    x=0;
    curr=1;
    l=strlen(temp);
    /* scan the string for '\n's, counting them. */
    while ((x<l)&&(curr<line_num))
        if (temp[x++]=='\n') curr++;
    /* set cursor position to beginning of the correct line. */
    XmTextSetInsertionPosition(text,(XmTextPosition)x);
    /* prevent memory leaks. */
    if (temp != NULL)
        XtFree(temp);
}

void jumpCB(w, client_data, call_data)
    Widget w;
    int client_data;
    XmSelectionBoxCallbackStruct *call_data;
/* handles jump-to-line dialog box. */
/* See Chapter 10. */
{
    char *jumpstr;

    if (client_data==CANCEL)
    {
        XtUnmanageChild (jump_dialog);
        return;
    }
        
    /* get the number entered */
    XmStringGetLtoR(call_data->value, char_set, &jumpstr);

    /* use the jump string to jump to a line */
    jump_to_line(atoi(jumpstr));

    /* popdown the file selection box */
    XtUnmanageChild (jump_dialog);
    XtFree(jumpstr);
}

void closeCB(w, client_data, call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
/* handles the "OK to close" question. */
{
    if (client_data==OK)
      handle_close();
}

char *string_search(cs,ct)
    char *cs;
    char *ct;
/* searches for ct in cs. Returns a pointer to the beginning of the
   first instance of ct. */
/* See Chapter 14. */
{
    int done;
    char *ct2,*cs2;

    /* check for "no work" situations */
    if (cs==NULL || ct==NULL)
        return NULL;
    if (*cs=='\0' || *ct=='\0')
        return NULL;

    /* loop through each character of cs. */
    done=False;
    while ((!done)&&(*cs!='\0'))
    {
        /* check to see if the first char of ct is in *cs. If it
           is proceed to check the rest of the letters in ct against
           cs. */
        if (*cs!=*ct)
            cs++;
        else
        {
            cs2=cs;
            ct2=ct;
            do
            {
                ct2++;
                cs2++;
            } while ((*cs2==*ct2) && (*ct2!='\0') && (*cs2!='\0'));
            if (*ct2=='\0')
            {
                done=True;
                return cs;
            }
            else
                cs++;
        }
    }
    if (!done)
        return NULL;
}

void lowercase(s)
    char *s;
/* converts s to lower case. */
/* See Chapter 14. */
{
    int x,y;

    y=strlen(s);
    for (x=0; x<y; x++)
    {
        if (s[x]>='A' && s[x]<='Z')
            s[x]=s[x]+32;
    }
}

void do_find()
/* finds the string in find_edit1 in the text starting at the current
   cursor position. */
/* See Chapter 14. */
{
    Arg al[10];
    int ac;
    XmTextPosition cursor_pos;
    char *find_string,*start,*temp,*p;
    Boolean found=False;
    int i;

    /* get the strings from the dialog box and the main text widget. */
    find_string=XmTextGetString(find_edit1);
    cursor_pos=XmTextGetInsertionPosition(text);
    start=XmTextGetString(text);
    temp=start+cursor_pos+1;
    if (!case_matters)
    {
      lowercase(temp);
      lowercase(find_string);
    }
    p=string_search(temp,find_string);
    /* if not found, display an error. */
    if (p==NULL)
    {
        ac=0;
        XtSetArg(al[ac], XmNmessageString, XmStringCreateLtoR(
            "String not found between current\ncursor location and end.", 
            char_set));  ac++;
        XtSetValues(finderror_dialog,al,ac);
        XtManageChild(finderror_dialog);
    }
    /* if found, select the found string and scroll it to the top of 
       the window. */
    else if (p!=NULL)
    {
        i=p-start;
        XmTextSetSelection(text,(XmTextPosition)i,
            (XmTextPosition)(i+strlen(find_string)),CurrentTime);
        XmTextSetInsertionPosition(text,(XmTextPosition)i);
        XmTextSetTopCharacter(text,(XmTextPosition)i);
        found=True;
    }
    XtFree(start);
    XtFree(find_string);
}

void do_change()
/* changes the found string to the new value. */
/* See Chapter 14. */
{
    Arg al[10];
    int ac;
    XmTextPosition cursor_pos;
    char *start,*temp,*p,*find_string,*replace_string;

    find_string=XmTextGetString(find_edit1);
    replace_string=XmTextGetString(find_edit2);
    cursor_pos=XmTextGetInsertionPosition(text);
    start=XmTextGetString(text);
    temp=start+cursor_pos;
    if (!case_matters)
    {
        lowercase(temp);
        lowercase(find_string);
    }
    if ((find_string==NULL)||
        (strncmp(temp,find_string,strlen(find_string))!=0))
    {
        ac = 0;
        XtSetArg(al[ac], XmNmessageString, XmStringCreateLtoR(
           "Change must be preceeded by a find.", 
           char_set));  ac++;
        XtSetValues(finderror_dialog, al, ac);
        XtManageChild(finderror_dialog);
    }
    else
    {
        XmTextReplace(text,cursor_pos,cursor_pos+
            (XmTextPosition)strlen(find_string),replace_string);
        XmTextSetSelection(text,cursor_pos,
            cursor_pos+(XmTextPosition)strlen(replace_string),CurrentTime);
        XmTextSetInsertionPosition(text,(XmTextPosition)(cursor_pos+
            strlen(replace_string)));
    }
    XtFree(find_string);
    XtFree(replace_string);
}

void do_find_change()
{
    do_change();
    do_find();
}

void do_case_sensitivity()
/* get the new value of the case toggle button. */
/* See Chapter 14. */
{
    Arg al[10];
    int ac;

    ac=0;
    XtSetArg(al[ac],XmNset,&case_matters); ac++;
    XtGetValues(case_toggle,al,ac);
}

void findCB(w,client_data,call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
/* callback for any button in the find dialog box. */
/* See Chapter 14. */
{
    switch (client_data)
    {
        case FIND_FIND:
            do_find();
            break;
        case FIND_FIND_CHANGE:
            do_find_change();
            break;
        case FIND_CHANGE:
            do_change();
            break;
        case FIND_CANCEL:
            XtUnmanageChild(find_dialog);
            break;
        case FIND_CASE:
            do_case_sensitivity();
            break;
    }
}

void quitCB(w, client_data, call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
/* handles the "OK to quit" question. */
{
    if (client_data==OK)
      exit(0);
}

void unmanageCB(w, client_data, call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
/* handles the unmanagement of any dialog needing to simply dissappear. */
{
    XtUnmanageChild(w);
}

void readonlyCB(w, client_data, call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
/* handles menu options. */
{
    XtSetSensitive(save_option,False);
    XtSetSensitive(cut_option,False);
    XtSetSensitive(clear_option,False);
    XtSetSensitive(paste_option,False);
    XtSetSensitive(find_change_button,False);
    XtSetSensitive(change_button,False);
    XmTextSetEditable(text,False);
    XtUnmanageChild(readonly_dialog);
}

void menuCB(w, client_data, call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
/* handles menu options. */
/* See Chapter 6, 10. */
{
    Time time;
    Arg al[10];
    int ac;
    XmString directory,dir_mask,pattern;
    char *s;

    switch (client_data)
    {
        case OPEN:
            /* make the file selection box appear. The DoSearch
               function is called each time so that new files are
               incorporated into the list of available files.*/
            ac=0;
            XtSetArg(al[ac],XmNdirMask,&dir_mask); ac++;
            XtGetValues(open_dialog,al,ac);

            XmFileSelectionDoSearch(open_dialog,dir_mask);

            XtManageChild(open_dialog);
            break;
        case CLOSE:
            if (text_changed) 
                XtManageChild(close_dialog);
            else
                handle_close();
            break;
        case SAVE:
            handle_save(filename);
            XtSetSensitive(save_option,False);  /* will get set true if  */	
                                                /* file is changed again.*/
            break;
        case SAVE_AS:
            /* get the directory from the open dialog filter so user knows
               where the filter path is pointing when using save_as. */
            ac=0;
            XtSetArg(al[ac],XmNdirectory,&directory); ac++;
            XtGetValues(open_dialog,al,ac);
            ac=0;
            XtSetArg(al[ac],XmNtextString,directory); ac++;
            XtSetValues(save_as_dialog,al,ac);
            XtFree(directory);
            XtManageChild(save_as_dialog);
            break;
        case NEW:
            /* get the directory from the open dialog filter so user knows
               where the filter path is pointing when using new. */
            ac=0;
            XtSetArg(al[ac],XmNdirectory,&directory); ac++;
            XtGetValues(open_dialog,al,ac);
            ac=0;
            XtSetArg(al[ac],XmNtextString,directory); ac++;
            XtSetValues(new_dialog,al,ac);
            XtFree(directory);
            XtManageChild(new_dialog);
            break;
        case QUIT:
            if (text_changed)
                XtManageChild(quit_dialog);
            else
                exit(0);
            break;

        case CUT:
            time=call_data->event->xbutton.time;
            XmTextCut(text,time);
            break;
        case CLEAR:
            XmTextRemove(text);
            break;
        case PASTE:
            XmTextPaste(text);
            break;
        case COPY:
            time=call_data->event->xbutton.time;
            XmTextCopy(text,time);
            break;

        case TOP:
            /* if the cursor is already at 0 and you have scrolled
               elsewhere with the scroll bar and then select top, the
               fact that you move the cursor from 0 to 0 will not
               cause the screen to update to the top of the file.
               To solve this problem, move to position 1 and then
               back to 0. Same technique is used in the other options.*/
            XmTextSetInsertionPosition(text,(XmTextPosition)1);
            XmTextSetInsertionPosition(text,(XmTextPosition)0);
            break;
        case BOTTOM:
        {
            XmTextPosition y,z;
            char *temp;

            temp=XmTextGetString(text);
            y=z=XmTextGetLastPosition(text);
            if (y!=0) z--;
            XmTextSetInsertionPosition(text,z);
            XmTextSetInsertionPosition(text,y);
            if (temp != NULL)
               XtFree(temp);
            break;
        }
        case JUMP_CURS:
        {
            XmTextPosition x,cursorPos;

            cursorPos=XmTextGetInsertionPosition(text);
            if (cursorPos>0) x=cursorPos-1; else x=cursorPos+1;
            XmTextSetInsertionPosition(text,x);
            XmTextSetInsertionPosition(text,cursorPos);
            break;
        }
        case JUMP_LINE:
            XtManageChild (jump_dialog);
            break;

        case ABOUT:
            XtManageChild (about_dialog);
            break;
        case HELP:
            XtManageChild (help_dialog);
            break;

        case FIND:
            XtManageChild(find_dialog);
            break;
    }
}

Widget make_help_menu(menu_name, mnemonic, menu_bar)
    char *menu_name;
    KeySym mnemonic;
    Widget menu_bar;
/* Creates a new menu on the menu bar. */
/* See Chapter 6. */
{
    int ac;
    Arg al[10];
    Widget menu, cascade;

    ac = 0;
    menu = XmCreatePulldownMenu (menu_bar, menu_name, al, ac);

    ac = 0;
    XtSetArg (al[ac], XmNsubMenuId, menu); ac++;
    XtSetArg (al[ac], XmNmnemonic, mnemonic); ac++;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR(menu_name, char_set)); ac++;
    cascade = XmCreateCascadeButton (menu_bar, menu_name, al, ac);
    XtManageChild (cascade); 

    ac=0;
    XtSetArg(al[ac],XmNmenuHelpWidget,cascade); ac++;
    XtSetValues(menu_bar,al,ac);

    return(menu);
}

void add_accelerator(w, acc_text, key)
    Widget w;
    char *acc_text;
    char *key;
/* adds an accelerator to a menu option. */
/* See Chapter 6. */
{
    int ac;
    Arg al[10];

    ac=0;
    XtSetArg(al[ac],XmNacceleratorText,
        XmStringCreate(acc_text,char_set)); ac++;
    XtSetArg(al[ac],XmNaccelerator,key); ac++;
    XtSetValues(w,al,ac);
}

Widget make_menu_option(option_name, mnemonic, client_data, menu)
    char *option_name;
    KeySym mnemonic;
    int client_data;
    Widget menu;
/* Adds an option to an existing menu. */
/* See Chapter 6. */
{
    int ac;
    Arg al[10];
    Widget b;

    ac = 0;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR(option_name,
        char_set)); ac++;
    XtSetArg (al[ac], XmNmnemonic, mnemonic); ac++;
    b=XtCreateManagedWidget(option_name,xmPushButtonWidgetClass,
        menu,al,ac);
    XtAddCallback (b, XmNactivateCallback, menuCB, client_data);
    return(b);
}

Widget make_menu(menu_name, mnemonic, menu_bar)
    char *menu_name;
    KeySym mnemonic;
    Widget menu_bar;
/* Creates a new menu on the menu bar. */
/* See Chapter 6. */
{
    int ac;
    Arg al[10];
    Widget menu, cascade;

    ac = 0;
    menu = XmCreatePulldownMenu (menu_bar, menu_name, al, ac);

    ac = 0;
    XtSetArg (al[ac], XmNsubMenuId, menu); ac++;
    XtSetArg (al[ac], XmNmnemonic, mnemonic); ac++;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR(menu_name, char_set)); ac++;
    cascade = XmCreateCascadeButton (menu_bar, menu_name, al, ac);
    XtManageChild (cascade); 

    return(menu);
}

void create_menus(menu_bar)
    Widget menu_bar;
/* See Chapter 6. */
{
    int ac;
    Arg al[10];
    Widget menu;

    menu=make_menu("File",'F',menu_bar);
    open_option = make_menu_option("Open",'O',OPEN,menu);
    add_accelerator(open_option,"meta+o","Meta<Key>o:");
    new_option = make_menu_option("New",'N',NEW,menu);
    add_accelerator(new_option,"meta+w","Meta<Key>w:");
    save_option = make_menu_option("Save",'S',SAVE,menu);
    add_accelerator(save_option,"meta+s","Meta<Key>s:");
    save_as_option = make_menu_option("Save As",'A',SAVE_AS,menu);
    close_option = make_menu_option("Close",'C',CLOSE,menu);
    add_accelerator(close_option,"meta+l","Meta<Key>l:");
    quit_option = make_menu_option("Exit",'E',QUIT,menu);
    add_accelerator(quit_option,"meta+q","Meta<Key>q:");

    menu=make_menu("Edit",'E',menu_bar);
    cut_option = make_menu_option("Cut",'C',CUT,menu);
    /* my Mac heritage is showing through on these accelerators. 
       Change them to something else if you want. */
    add_accelerator(cut_option,"meta+x","Meta<Key>x:");
    copy_option = make_menu_option("Copy",'o',COPY,menu);
    add_accelerator(copy_option,"meta+c","Meta<Key>c:");
    paste_option = make_menu_option("Paste",'P',PASTE,menu);
    add_accelerator(paste_option,"meta+v","Meta<Key>v:");
    clear_option = make_menu_option("Clear",'r',CLEAR,menu);

    menu=make_menu("Navigate",'N',menu_bar);
    top_option = make_menu_option("Top",'T',TOP,menu);
    add_accelerator(top_option,"meta+t","Meta<Key>t:");
    bottom_option = make_menu_option("Bottom",'B',BOTTOM,menu);
    add_accelerator(bottom_option,"meta+b","Meta<Key>b:");
    jump_curs_option = make_menu_option("Jump to Cursor",'C',JUMP_CURS,menu);
    jump_line_option = make_menu_option("Jump to Line",'L',JUMP_LINE,menu);
    add_accelerator(jump_line_option,"meta+j","Meta<Key>j:");

    menu=make_help_menu("Help",'H',menu_bar);
    about_option=make_menu_option("About",'A',ABOUT,menu);
    help_option=make_menu_option("Help",'H',HELP,menu);

    menu=make_menu("Utilities",'U',menu_bar);
    find_option = make_menu_option("Find/Change",'F',FIND,menu);

    change_menu_sensitivity(True);
}

void create_find_dialog()
/* creates all of the widgets in the find dialog box. */
/* See Chapter 13. */
{
    Arg al[10];
    int ac;

    /* create but do NOT manage the container dialog. */
    ac=0;
    XtSetArg(al[ac],XmNheight,200); ac++;
    XtSetArg(al[ac],XmNwidth,400); ac++;
    XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
    XtSetArg(al[ac],XmNnoResize,True); ac++;
    XtSetArg(al[ac],XmNdialogStyle,XmDIALOG_MODELESS); ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Find",char_set)); ac++;
    find_dialog=XmCreateBulletinBoardDialog(toplevel,
        "find_dialog",al,ac);

    /* create and manage the two labels. */
    ac=0;
    XtSetArg(al[ac],XmNx,10); ac++;
    XtSetArg(al[ac],XmNy,10); ac++;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR("Find:", char_set));    ac++;
    find_label1=XmCreateLabel(find_dialog,"find_label1",al,ac);
    XtManageChild(find_label1);
     
    ac=0;
    XtSetArg(al[ac],XmNx,10); ac++;
    XtSetArg(al[ac],XmNy,50); ac++;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR("Change to:", char_set));    ac++;
    find_label2=XmCreateLabel(find_dialog,"find_label2",al,ac);
    XtManageChild(find_label2);
     
    /* Create and manage the two text widgets. */
    ac=0;
    XtSetArg(al[ac],XmNx,100); ac++;
    XtSetArg(al[ac],XmNy,10); ac++;
    find_edit1=XmCreateText(find_dialog,"find_edit1",al,ac);
    XtManageChild(find_edit1);
     
    ac=0;
    XtSetArg(al[ac],XmNx,100); ac++;
    XtSetArg(al[ac],XmNy,50); ac++;
    find_edit2=XmCreateText(find_dialog,"find_edit2",al,ac);
    XtManageChild(find_edit2);
     
    /* create and manage the four pushbuttons inside a rowcolumn widget. */
    ac=0;
    XtSetArg(al[ac],XmNx,0); ac++;
    XtSetArg(al[ac],XmNy,90); ac++;
    XtSetArg(al[ac],XmNorientation,XmHORIZONTAL); ac++;
    XtSetArg(al[ac],XmNpacking,XmPACK_TIGHT); ac++;
    XtSetArg(al[ac],XmNadjustLast,False); ac++;
    find_rc=XmCreateRowColumn(find_dialog,"find_rc",al,ac);
    XtManageChild(find_rc);

    ac=0;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR("Find", char_set));    ac++;
    find_button=XmCreatePushButton(find_rc,"find_button",al,ac);
    XtManageChild(find_button);
    XtAddCallback (find_button, XmNactivateCallback, findCB, FIND_FIND);
     
    ac=0;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR("Change, then Find", char_set));    ac++;
    find_change_button=XmCreatePushButton(find_rc,"find_change_button",
        al,ac);
    XtManageChild(find_change_button);
    XtAddCallback (find_change_button, XmNactivateCallback, findCB, 
        FIND_FIND_CHANGE);

    ac=0;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR("Change", char_set));    ac++;
    change_button=XmCreatePushButton(find_rc,"change_button",al,ac);
    XtManageChild(change_button);
    XtAddCallback (change_button, XmNactivateCallback, findCB, 
        FIND_CHANGE);

    ac=0;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR("Top", char_set));    ac++;
    find_top_button=XmCreatePushButton(find_rc,"find_top_button",al,ac);
    XtManageChild(find_top_button);
    XtAddCallback (find_top_button, XmNactivateCallback, menuCB, 
        TOP);

    ac=0;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR("Cancel", char_set));    ac++;
    cancel_button=XmCreatePushButton(find_rc,"cancel_button",al,ac);
    XtManageChild(cancel_button);
    XtAddCallback (cancel_button, XmNactivateCallback, findCB, 
        FIND_CANCEL);

    /* create and manage the toggle button. */
    ac=0;
    XtSetArg(al[ac],XmNx,10); ac++;
    XtSetArg(al[ac],XmNy,130); ac++;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR("Case Matters", char_set));    ac++;
    case_toggle=XmCreateToggleButton(find_dialog,"case_toggle",al,ac);
    XtManageChild(case_toggle);
    XtAddCallback (case_toggle, XmNvalueChangedCallback, findCB, 
        FIND_CASE);

    /* set the default and cancel button for the find dialog. */
    ac=0;
    XtSetArg(al[ac],XmNdefaultButton,find_button); ac++;
    XtSetArg(al[ac],XmNcancelButton,cancel_button); ac++;
    XtSetValues(find_dialog,al,ac);
}

void create_dialog_boxes()
/* See Chapter 7. */
{
    Arg al[10];
    int ac;

    /* create the file selection box used by open option. */
    ac=0;
    XtSetArg(al[ac],XmNmustMatch,True); ac++;
    XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Open",char_set)); ac++;
    open_dialog=XmCreateFileSelectionDialog(toplevel,"open_dialog",al,ac);
    XtAddCallback (open_dialog, XmNokCallback, openCB, OK);
    XtAddCallback (open_dialog, XmNcancelCallback, openCB, CANCEL);
    XtUnmanageChild(XmSelectionBoxGetChild(open_dialog,
        XmDIALOG_HELP_BUTTON));

    /* create the new file prompt dialog. */
    ac = 0;
    XtSetArg(al[ac], XmNselectionLabelString, XmStringCreateLtoR
       ("Enter the name of the new file.", char_set));  ac++;
    XtSetArg(al[ac],XmNdialogStyle,XmDIALOG_APPLICATION_MODAL); ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: New",char_set)); ac++;
    new_dialog = XmCreatePromptDialog(toplevel,
       "new_dialog", al, ac);
    XtAddCallback (new_dialog, XmNokCallback, openCB, OK);
    XtAddCallback (new_dialog, XmNcancelCallback, openCB, CANCEL);
    XtUnmanageChild(XmSelectionBoxGetChild(new_dialog,
        XmDIALOG_HELP_BUTTON));

    /* create the overwrite error dialog. */
    ac = 0;
    XtSetArg(al[ac], XmNmessageString, XmStringCreateLtoR
        ("The file exists.\nOK to overwrite?", 
        char_set));  ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Overwrite",char_set)); ac++;
    overwrite_dialog = XmCreateQuestionDialog(toplevel,
       "overwrite_dialog", al, ac);
    XtAddCallback (overwrite_dialog, XmNokCallback, overwriteCB, OK);
    XtAddCallback (overwrite_dialog, XmNcancelCallback, overwriteCB, CANCEL);
    XtUnmanageChild(XmMessageBoxGetChild(overwrite_dialog,
        XmDIALOG_HELP_BUTTON));

    /* create the jump prompt dialog. */
    ac = 0;
    XtSetArg(al[ac], XmNselectionLabelString, XmStringCreateLtoR
       ("Enter line number to jump to.", char_set));  ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Jump to Line",char_set)); ac++;
    jump_dialog = XmCreatePromptDialog(toplevel, "jump_dialog", al, ac);
    XtUnmanageChild(XmSelectionBoxGetChild(jump_dialog,
        XmDIALOG_HELP_BUTTON));
    XtAddCallback (jump_dialog, XmNokCallback, jumpCB, OK);
    XtAddCallback (jump_dialog, XmNcancelCallback, jumpCB, CANCEL);

    /* create the save_as prompt dialog. */
    ac = 0;
    XtSetArg(al[ac], XmNselectionLabelString, XmStringCreateLtoR
       ("Enter the new file name to save the file as.", 
       char_set));  ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Save As",char_set)); ac++;
    save_as_dialog = XmCreatePromptDialog(toplevel, "save_as_dialog", al, ac);
    XtUnmanageChild(XmSelectionBoxGetChild(save_as_dialog,
        XmDIALOG_HELP_BUTTON));
    XtAddCallback (save_as_dialog, XmNokCallback, save_asCB, OK);
    XtAddCallback (save_as_dialog, XmNcancelCallback, save_asCB, CANCEL);

    /* create the quit question dialog. */
    ac = 0;
    XtSetArg(al[ac], XmNmessageString, XmStringCreateLtoR
       ("The file has been changed.\nOK to quit?", char_set));  ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Quit Check",char_set)); ac++;
    quit_dialog = XmCreateQuestionDialog(toplevel, "quit_dialog", 
        al, ac);
    XtUnmanageChild(XmMessageBoxGetChild(quit_dialog,
        XmDIALOG_HELP_BUTTON));
    XtAddCallback (quit_dialog, XmNokCallback, quitCB, OK);
    XtAddCallback (quit_dialog, XmNcancelCallback, quitCB, CANCEL);

    /* create the close question dialog. */
    ac = 0;
    XtSetArg(al[ac], XmNmessageString, XmStringCreateLtoR
       ("The file has been changed.\nOK to close without saving?", 
       char_set));  ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Close Check",char_set)); ac++;
    close_dialog = XmCreateQuestionDialog(toplevel, "close_dialog", 
        al, ac);
    XtUnmanageChild(XmMessageBoxGetChild(close_dialog,
        XmDIALOG_HELP_BUTTON));
    XtAddCallback (close_dialog, XmNokCallback, closeCB, OK);
    XtAddCallback (close_dialog, XmNcancelCallback, closeCB, CANCEL);

    /* create the read only dialog. */
    ac = 0;
    XtSetArg(al[ac], XmNmessageString, XmStringCreateLtoR
       ("This file is read-only.", char_set));  ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Read Only",char_set)); ac++;
    readonly_dialog = XmCreateInformationDialog(toplevel, "readonly_dialog", 
        al, ac);
    XtUnmanageChild(XmMessageBoxGetChild(readonly_dialog,
        XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(readonly_dialog,
        XmDIALOG_CANCEL_BUTTON));
    XtAddCallback (readonly_dialog, XmNokCallback, readonlyCB, OK);

    /* create the write error dialog. */
    ac = 0;
    XtSetArg(al[ac], XmNmessageString, XmStringCreateLtoR
       ("The file could not be written.", char_set));  ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Write Error",char_set)); ac++;
    error_dialog = XmCreateErrorDialog(toplevel, "error_dialog", 
        al, ac);
    XtUnmanageChild(XmMessageBoxGetChild(error_dialog,
        XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(error_dialog,
        XmDIALOG_CANCEL_BUTTON));
    XtAddCallback (error_dialog, XmNokCallback, unmanageCB, OK);

    /* create the help dialog. */
    ac = 0;
    XtSetArg(al[ac], XmNmessageString, XmStringCreateLtoR
        ("This is a fairly simple program, and should be self-explanatory.\nThe only question I've had is, 'What does Jump to Cursor do?'\nIf you have used the scroll bar to move around in the document,\nthen Jump to Cursor will take you back to the current cursor\nposition and display it.\n   Occasionally the scroll bar in the text widget will not allow\nyou to scroll all the way to the bottom of the text. I believe\nthis is a problem inside the text widget. To solve this, use\nthe Navigate/Bottom menu option. It will go to the bottom of\nthe text and reset the scroll bar.", 
        char_set));  ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Help",char_set)); ac++;
    help_dialog = XmCreateInformationDialog(toplevel, "help_dialog", 
        al, ac);
    XtUnmanageChild(XmMessageBoxGetChild(help_dialog,
        XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(help_dialog,
        XmDIALOG_CANCEL_BUTTON));
    XtAddCallback (help_dialog, XmNokCallback, unmanageCB, OK);

    /* create the find error dialog. */
    ac=0;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: Find error",char_set)); ac++;
    finderror_dialog = XmCreateMessageDialog(find_dialog,
       "fr_error", al, ac);
    XtAddCallback (finderror_dialog, XmNokCallback,
        unmanageCB, OK);
    XtUnmanageChild(XmMessageBoxGetChild(finderror_dialog,
        XmDIALOG_CANCEL_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(finderror_dialog,
        XmDIALOG_HELP_BUTTON));

    /* create the about box dialog. */
    ac = 0;
    XtSetArg(al[ac], XmNmessageAlignment, XmALIGNMENT_CENTER); ac++;
    XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR(
        "NMG Editor: About",char_set)); ac++;
    XtSetArg(al[ac], XmNmessageString, XmStringCreateLtoR
       ("NMG Editor\n\nPresented by the NCSU Motif Group\n\nVersion 1.03, released 11/5/91\nby Marshall Brain.\nemail: brain@eos.ncsu.edu", 
       char_set));  ac++;
    about_dialog = XmCreateMessageDialog(toplevel, "about_dialog", 
        al, ac);
    XtUnmanageChild(XmMessageBoxGetChild(about_dialog,
        XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(about_dialog,
        XmDIALOG_CANCEL_BUTTON));
    XtAddCallback (about_dialog, XmNokCallback, unmanageCB, OK);
}

void create_icon()
/* creates the icon pixmap and title. */
/* See Chapter 14. */
{
    Pixmap p;
    Arg al[10];
    int ac;

    p=XCreateBitmapFromData(XtDisplay(toplevel),
        RootWindowOfScreen(XtScreen(toplevel)), 
        nmg_bits, nmg_width, nmg_height);
    ac=0;
    XtSetArg(al[ac], XmNiconPixmap, p); ac++;
    XtSetArg(al[ac], XmNiconName, "Editor"); ac++;
    XtSetValues(toplevel,al,ac);
}

void main(argc, argv)
    int argc;
    char *argv[];
/* See Chapter 2, 5, 10. */
{
    Arg al[10];
    int ac;
    XFontStruct *font=NULL;
    XmFontList fontlist=NULL;

    toplevel=XtAppInitialize(&context,"",NULL,0,&argc,argv,NULL,NULL,0);

    /* default window size and title. */
    ac=0;
    XtSetArg(al[ac], XmNtitle, "NMG Editor"); ac++;
    XtSetArg(al[ac],XmNheight,400); ac++;
    XtSetArg(al[ac],XmNwidth,600); ac++;
    XtSetValues(toplevel,al,ac);

    /* create a form widget. */
    ac=0;
    form=XtCreateManagedWidget("form",xmFormWidgetClass,
        toplevel,al,ac);

    /* create a menu bar and attach it to the form. */
    ac=0;
    XtSetArg(al[ac], XmNtopAttachment,   XmATTACH_FORM); ac++;
    XtSetArg(al[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    XtSetArg(al[ac], XmNleftAttachment,  XmATTACH_FORM); ac++;
    menu_bar=XmCreateMenuBar(form,"menu_bar",al,ac);
    XtManageChild(menu_bar);

    /* create a text widget and attach it to the form. */
    ac=0;
    XtSetArg(al[ac], XmNtopAttachment,   XmATTACH_WIDGET); ac++;
    XtSetArg(al[ac], XmNtopWidget, menu_bar); ac++;
    XtSetArg(al[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    XtSetArg(al[ac], XmNleftAttachment,  XmATTACH_FORM); ac++;
    XtSetArg(al[ac], XmNbottomAttachment,XmATTACH_FORM); ac++;
    XtSetArg(al[ac],XmNeditMode,XmMULTI_LINE_EDIT); ac++;
    text=XmCreateScrolledText(form, "text", al, ac);
    XtManageChild(text);
    XtSetSensitive(text,False);
    XmTextSetEditable(text,False);
    XtAddCallback(text, XmNvalueChangedCallback, text_changedCB, NULL);

    ac=0;
    font=XLoadQueryFont(XtDisplay(text),"fixed");
    fontlist=XmFontListCreate(font,char_set);
    XtSetArg(al[ac],XmNfontList,fontlist); ac++;
    XtSetValues(text,al,ac);

    create_find_dialog();
    create_dialog_boxes();
    create_menus(menu_bar);
    create_icon();

    XtRealizeWidget(toplevel);
    XtAppMainLoop(context);
}

