/***************************************************************
From the book "Motif Programming, The Essentials...and More"
by Marshall Brain.
Published by Digital Press, ISBN 1-55558-089-0.
To order the book, call 1-800-DIGITAL and ask for EY-J816E-DP.

Copyright 1992, by Digital Equipment Corp.

This code implements rubber-banding in a drawing area.
***************************************************************/

/* rubber_band.c */

#include <Xm/Xm.h>
#include <Xm/DrawingA.h>

XtAppContext context;

Widget toplevel;
Widget drawing_area;
GC gc_copy;
GC gc_xor;
int start_x,start_y;
int old_x,old_y;

void setup_gcs()
/* set up the graphics context. */
{
    int foreground,background;
    Arg al[10];
    int ac;
    XGCValues vals;

    /* get the current fg and bg colors. */
    ac=0;
    XtSetArg(al[ac], XmNforeground, &foreground); ac++;
    XtSetArg(al[ac], XmNbackground, &background); ac++;
    XtGetValues(drawing_area, al, ac);

    /* create the copy gc. */
    vals.foreground = foreground;
    vals.background = background;
    gc_copy= XtGetGC(drawing_area, GCForeground | GCBackground, &vals);

    /* create the xor gc. */
    vals.foreground = foreground ^ background;
    vals.function = GXxor;
    gc_xor= XtGetGC(drawing_area, GCForeground | GCBackground |
        GCFunction, &vals);
}

void handle_start(w,client_data,event)
    Widget w;
    XtPointer client_data;
    XEvent *event;
/* handles the ButtonPress event */
{
    old_x=start_x=event->xbutton.x;
    old_y=start_y=event->xbutton.y;
    XDrawLine(XtDisplay(w), XtWindow(w), gc_xor, 
        start_x, start_y, old_x, old_y);
}

void handle_drag(w,client_data,event)
    Widget w;
    XtPointer client_data;
    XEvent *event;
/* handles the ButtonMotion event */
{
    XDrawLine(XtDisplay(w), XtWindow(w), gc_xor, 
        start_x, start_y, old_x, old_y);
    old_x=event->xbutton.x;
    old_y=event->xbutton.y;
    XDrawLine(XtDisplay(w), XtWindow(w), gc_xor, 
        start_x, start_y, old_x, old_y);
}

void handle_done(w,client_data,event)
    Widget w;
    XtPointer client_data;
    XEvent *event;
/* handles the ButtonRelease event */
{
    old_x=event->xbutton.x;
    old_y=event->xbutton.y;
    XDrawLine(XtDisplay(w), XtWindow(w), gc_copy, 
        start_x, start_y, old_x, old_y);
}

void main(argc,argv)
    int argc;
    char *argv[];
{
    Arg al[10];
    int ac;

    /* create the toplevel shell */
    toplevel = XtAppInitialize(&context,"",NULL,0,&argc,argv,
        NULL,NULL,0);

    /* default window size. */
    ac=0;
    XtSetArg(al[ac],XmNheight,400); ac++;
    XtSetArg(al[ac],XmNwidth,400); ac++;
    XtSetValues(toplevel,al,ac);

    /* create drawing area */
    ac=0;
    drawing_area=XmCreateDrawingArea(toplevel,"drawing_area",al,ac);
    XtManageChild(drawing_area);
    /* add in event handlers for drawing area */
    XtAddEventHandler(drawing_area, ButtonPressMask, FALSE,
        handle_start, NULL);
    XtAddEventHandler(drawing_area, ButtonReleaseMask, FALSE,
        handle_done, NULL);
    XtAddEventHandler(drawing_area, ButtonMotionMask, FALSE,
        handle_drag, NULL);

    setup_gcs();

    XtRealizeWidget(toplevel);
    XtAppMainLoop(context);
}
