/***************************************************************
From the book "Motif Programming, The Essentials...and More"
by Marshall Brain.
Published by Digital Press, ISBN 1-55558-089-0.
To order the book, call 1-800-DIGITAL and ask for EY-J816E-DP.

Copyright 1992, by Digital Equipment Corp.

This code demonstrates the creation of a large virtual drawing
area controlled by a vertical scroll bar.
***************************************************************/

/* virtual.c*/

#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/ScrollBar.h>

#define MAX_Y 5000

XtAppContext context;

GC gc_copy;
GC gc_xor;
Widget toplevel, drawing_area, form, scroll;
int start_x,start_y;
int old_x,old_y;
int y_offset=0;

/* node for the single linked list contains the 4 line coordinates 
     and a next pointer. */
struct node 
{
    int x1,y1,x2,y2;
    struct node *next;
};

/* first points to the first node in the SLL. */
struct node *first=NULL;

void setup_gcs()
/* set up the graphics contexts */
{
    int foreground,background;
    Arg al[10];
    int ac;
    XGCValues vals;

    /* get the current fg and bg colors. */
    ac=0;
    XtSetArg(al[ac], XmNforeground, &foreground); ac++;
    XtSetArg(al[ac], XmNbackground, &background); ac++;
    XtGetValues(drawing_area, al, ac);

    /* create the copy gc. */
    vals.foreground = foreground;
    vals.background = background;
    gc_copy= XtGetGC(drawing_area, GCForeground | GCBackground, &vals);

    /* create the xor gc. */
    vals.foreground = foreground ^ background;
    vals.function = GXxor;
    gc_xor= XtGetGC(drawing_area, GCForeground | GCBackground |
                GCFunction, &vals);
}

void handle_start(w,client_data,event)
    Widget w;
    XtPointer client_data;
    XEvent *event;
/* See Section 17.18. */
{
    old_x=start_x=event->xbutton.x;
    old_y=start_y=event->xbutton.y;
    XDrawLine(XtDisplay(w), XtWindow(w), gc_xor, 
        start_x, start_y, old_x, old_y);
}

void handle_drag(w,client_data,event)
    Widget w;
    XtPointer client_data;
    XEvent *event;
/* See Section 17.18. */
{
    XDrawLine(XtDisplay(w), XtWindow(w), gc_xor, 
        start_x, start_y, old_x, old_y);
    old_x=event->xbutton.x;
    old_y=event->xbutton.y;
    XDrawLine(XtDisplay(w), XtWindow(w), gc_xor, 
        start_x, start_y, old_x, old_y);
}

void handle_done(w,client_data,event)
    Widget w;
    XtPointer client_data;
    XEvent *event;
/* See Section 17.18. */
{
    struct node *temp;

    old_x=event->xbutton.x;
    old_y=event->xbutton.y;
    XDrawLine(XtDisplay(w), XtWindow(w), gc_copy, 
        start_x, start_y, old_x, old_y);

    /* add the new line to the SLL. Add y_offset to y coords so that
         lines in the SLL are in virtual space coordinates. */
    temp=(struct node *) malloc(sizeof(struct node));
    temp->x1=start_x;    
    temp->y1=start_y+y_offset;    
    temp->x2=old_x;    
    temp->y2=old_y+y_offset;    
    temp->next=first;
    first=temp;
}

void exposeCB(w,client_data,call_data)
    Widget w;
    XtPointer client_data;
    XtPointer call_data;
/* called whenever drawing area is exposed. */
{
    struct node *temp;

    /* Traverse the SLL and draw all lines to the drawing area. 
         Subtract off y_offset as each line is drawn so that virtual
         space is mapped correctly into the actual drawing area. */
    printf("expose event generated\n");
    temp=first;
    while (temp)
    {
        XDrawLine(XtDisplay(w), XtWindow(w), gc_copy, 
            temp->x1, temp->y1-y_offset, temp->x2, temp->y2-y_offset );
        temp=temp->next;
    }
}

void scrollCB(w,client_data,call_data)
    Widget w;
    XtPointer client_data;
    XtPointer call_data;
/* called whenever scrollbar moves. */
{
    Arg al[10];
    int ac;
    int value;

    XClearArea(XtDisplay(drawing_area), XtWindow(drawing_area), 
        0, 0, 0, 0, True); /* will generate a subsequent expose event. */

    /* get new scroll bar value and adjust y_offset. */
    ac=0;
    XtSetArg(al[ac],XmNvalue,&value); ac++;
    XtGetValues(w,al,ac);
    printf("scrollbar value = %d\n",value);

    y_offset=value;
}

void resizeCB(w,client_data,call_data)
    Widget w;
    XtPointer client_data;
    XtPointer call_data;
/* called whenever drawing area is resized. */
{
    Dimension wdth,hght;
    Arg al[10];
    int ac;

    /* get new window size. */
    ac=0;
    XtSetArg(al[ac],XmNheight,&hght); ac++;
    XtSetArg(al[ac],XmNwidth,&wdth); ac++;
    XtGetValues(w,al,ac);
    printf("resized to %dx%d\n",wdth,hght);

    /* adjust scroll bar. */
    ac=0;
    XtSetArg(al[ac],XmNsliderSize,hght); ac++;
    XtSetArg(al[ac],XmNpageIncrement,hght/2); ac++;
    XtSetValues(scroll,al,ac);
}

void main(argc,argv)
    int argc;
    char *argv[];
{
    Arg al[10];
    int ac;

    /* create the toplevel shell */
    toplevel = XtAppInitialize(&context,"",NULL,0,&argc,argv,
        NULL,NULL,0);

    /* default window size. */
    ac=0;
    XtSetArg(al[ac],XmNheight,400); ac++;
    XtSetArg(al[ac],XmNwidth,400); ac++;
    XtSetValues(toplevel,al,ac);

    /* create a form to hold widgets */
    ac=0;
    form=XmCreateForm(toplevel,"form",al,ac);
    XtManageChild(form);

    /* create a scroll bar */
    ac=0;
    XtSetArg(al[ac], XmNtopAttachment,        XmATTACH_FORM); ac++;
    XtSetArg(al[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
    XtSetArg(al[ac], XmNleftAttachment, XmATTACH_NONE); ac++;
    XtSetArg(al[ac], XmNbottomAttachment,     XmATTACH_FORM); ac++;
    XtSetArg(al[ac], XmNwidth, 20); ac++;
    XtSetArg(al[ac], XmNmaximum, MAX_Y); ac++;
    scroll=XmCreateScrollBar(form,"scroll",al,ac);
    XtManageChild(scroll);
    XtAddCallback(scroll,XmNvalueChangedCallback,scrollCB, NULL);

    /* create drawing area */
    ac=0;
    XtSetArg(al[ac], XmNtopAttachment,        XmATTACH_FORM); ac++;
    XtSetArg(al[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
    XtSetArg(al[ac], XmNrightAttachment, XmATTACH_WIDGET); ac++;
    XtSetArg(al[ac], XmNrightWidget, scroll); ac++;
    XtSetArg(al[ac], XmNbottomAttachment,     XmATTACH_FORM); ac++;
    drawing_area=XmCreateDrawingArea(form,"drawing_area",al,ac);
    XtManageChild(drawing_area);
    /* add in event handler */
    XtAddEventHandler(drawing_area, ButtonPressMask, FALSE,
         handle_start, NULL);
    XtAddEventHandler(drawing_area, ButtonReleaseMask, FALSE,
         handle_done, NULL);
    XtAddEventHandler(drawing_area, ButtonMotionMask, FALSE,
         handle_drag, NULL);
    XtAddCallback(drawing_area, XmNexposeCallback,
         exposeCB, NULL);
    XtAddCallback(drawing_area, XmNresizeCallback,
         resizeCB, NULL);

    setup_gcs();

    XtRealizeWidget(toplevel);
    XtAppMainLoop(context);
}
